"""Tests for Volvo Cars API."""

from functools import lru_cache
import json
import pathlib
from typing import Any
from unittest.mock import AsyncMock


def configure_mock(
    mock: AsyncMock, *, return_value: Any = None, side_effect: Any = None
) -> None:
    """Reconfigure mock."""
    mock.reset_mock()
    mock.side_effect = side_effect
    mock.return_value = return_value


@lru_cache
def load_json_fixture(name: str) -> dict:
    """Load a JSON object from a fixture."""

    name = f"{name}.json"
    fixtures_path = pathlib.Path().cwd().joinpath("tests", "fixtures")
    data_path = fixtures_path.joinpath(name)

    fixture = data_path.read_text(encoding="utf8")

    return json.loads(fixture)



"""Test fixtures."""

from unittest.mock import AsyncMock

from aiohttp import ClientSession
import pytest

from volvocarsapi.auth import VolvoCarsAuth


@pytest.fixture
def mock_client_session() -> AsyncMock:
    """Mock client session."""
    return AsyncMock(spec=ClientSession)


@pytest.fixture
def mock_token_manager() -> AsyncMock:
    """Mock token manager."""

    def _create_mock() -> VolvoCarsAuth:
        return AsyncMock(spec=VolvoCarsAuth)

    mock = _create_mock()
    mock.async_get_access_token = AsyncMock(return_value="valid_token_123")

    return mock



"""API tests."""

from datetime import UTC, datetime
from unittest.mock import AsyncMock

from aiohttp import ClientResponseError, ClientSession, RequestInfo
import pytest

from volvocarsapi.api import VolvoCarsApi
from volvocarsapi.auth import VolvoCarsAuth
from volvocarsapi.models import VolvoApiException, VolvoAuthException

from . import configure_mock, load_json_fixture


async def test_get_odometer_successful(
    mock_client_session: ClientSession, mock_token_manager: VolvoCarsAuth
) -> None:
    """Test successful odometer request."""
    odometer_json = load_json_fixture("odometer")
    _mock_response(mock_client_session, odometer_json)

    api = VolvoCarsApi(mock_client_session, mock_token_manager, "secretapikey")

    result = await api.async_get_odometer()
    assert result["odometer"].value == 30000
    assert result["odometer"].unit == "km"
    assert result["odometer"].timestamp == datetime(
        2024, 12, 30, 14, 18, 56, 0, tzinfo=UTC
    )


async def test_get_odometer_token_failure(
    mock_client_session: ClientSession, mock_token_manager: VolvoCarsAuth
) -> None:
    """Test token failure for odometer request."""
    configure_mock(
        mock_token_manager.async_get_access_token,
        side_effect=ClientResponseError(None, None, status=400),
    )

    api = VolvoCarsApi(mock_client_session, mock_token_manager, "secretapikey")

    with pytest.raises(VolvoAuthException):
        await api.async_get_odometer()


async def test_get_odometer_failure(
    mock_client_session: ClientSession, mock_token_manager: VolvoCarsAuth
) -> None:
    """Test general failure for odometer request."""
    _mock_response_failure(mock_client_session, 500)

    api = VolvoCarsApi(mock_client_session, mock_token_manager, "secretapikey")

    with pytest.raises(VolvoApiException):
        await api.async_get_odometer()


async def test_get_odometer_auth_failure(
    mock_client_session: ClientSession, mock_token_manager: VolvoCarsAuth
) -> None:
    """Test auth failure for odometer request."""
    _mock_response_failure(mock_client_session, 401)

    api = VolvoCarsApi(mock_client_session, mock_token_manager, "secretapikey")

    with pytest.raises(VolvoAuthException):
        await api.async_get_odometer()


async def test_access_token_successful(
    mock_client_session: ClientSession, mock_token_manager: VolvoCarsAuth
) -> None:
    """Test successful access token request."""
    api = VolvoCarsApi(mock_client_session, mock_token_manager, "secretapikey")

    access_token = await api.async_get_access_token()
    assert access_token == "valid_token_123"


@pytest.mark.parametrize(
    "status",
    [400, 401, 403],
)
async def test_access_token_raises_auth_exception(
    mock_client_session: ClientSession, mock_token_manager: VolvoCarsAuth, status: int
) -> None:
    """Test if getting access token raises a `VolvoAuthException`."""
    configure_mock(
        mock_token_manager.async_get_access_token,
        side_effect=ClientResponseError(None, None, status=status),
    )

    api = VolvoCarsApi(mock_client_session, mock_token_manager, "secretapikey")

    with pytest.raises(VolvoAuthException):
        await api.async_get_access_token()


@pytest.mark.parametrize(
    "status",
    [404, 500],
)
async def test_access_token_raises_api_exception(
    mock_client_session: ClientSession, mock_token_manager: VolvoCarsAuth, status: int
) -> None:
    """Test if getting access token raises a `VolvoApiException`."""
    configure_mock(
        mock_token_manager.async_get_access_token,
        side_effect=ClientResponseError(None, None, status=status),
    )

    api = VolvoCarsApi(mock_client_session, mock_token_manager, "secretapikey")

    with pytest.raises(VolvoApiException):
        await api.async_get_access_token()


def _mock_response(mock_client_session: ClientSession, json: dict) -> None:
    mock_response = AsyncMock()
    mock_response.__aenter__.return_value.json = AsyncMock(return_value=json)
    mock_client_session.request.return_value = mock_response


def _mock_response_failure(mock_client_session: ClientSession, status: int) -> None:
    mock_client_session.request.side_effect = ClientResponseError(
        RequestInfo(url="http://127.0.0.1", method="GET", headers={}),
        None,
        status=status,
    )



"""Volvo Cars API."""

from http import HTTPStatus
import logging
from typing import Any, cast

from aiohttp import (
    ClientError,
    ClientResponseError,
    ClientSession,
    ClientTimeout,
    RequestInfo,
    hdrs,
)
from yarl import URL

from .auth import AccessTokenManager
from .models import (
    VolvoApiException,
    VolvoAuthException,
    VolvoCarsAvailableCommand,
    VolvoCarsCommandResult,
    VolvoCarsErrorResult,
    VolvoCarsLocation,
    VolvoCarsValue,
    VolvoCarsValueField,
    VolvoCarsValueStatusField,
    VolvoCarsVehicle,
)
from .util import redact_data, redact_url

_API_CONNECTED_ENDPOINT = "/connected-vehicle/v2/vehicles"
_API_ENERGY_ENDPOINT = "/energy/v1/vehicles"
_API_ENERGY_V2_ENDPOINT = "/energy/v2/vehicles"
_API_LOCATION_ENDPOINT = "/location/v1/vehicles"
_API_URL = "https://api.volvocars.com"
_API_STATUS_URL = "https://public-developer-portal-bff.weu-prod.ecpaz.volvocars.biz/api/v1/backend-status"
_API_REQUEST_TIMEOUT = ClientTimeout(total=30)

_DATA_TO_REDACT = [
    "coordinates",
    "heading",
    "href",
    "vin",
]

_LOGGER = logging.getLogger(__name__)


class VolvoCarsApi:
    """Volvo Cars API."""

    def __init__(
        self,
        client: ClientSession,
        token_manager: AccessTokenManager,
        api_key: str,
        vin: str = "",
    ) -> None:
        """Initialize Volvo Cars API."""
        self._client = client
        self._token_manager = token_manager
        self.api_key = api_key
        self.vin = vin

    async def async_get_api_status(self) -> dict[str, VolvoCarsValue]:
        """Check the API status."""
        try:
            _LOGGER.debug("Request [API status]")
            async with self._client.get(
                _API_STATUS_URL, timeout=_API_REQUEST_TIMEOUT
            ) as response:
                _LOGGER.debug("Request [API status] status: %s", response.status)
                response.raise_for_status()
                json = await response.json()
                data = cast(dict[str, Any], json)
                _LOGGER.debug("Request [API status] response: %s", data)

                message = data.get("message") or "OK"

        except (ClientError, TimeoutError) as ex:
            _LOGGER.debug("Request [API status] error: %s", ex)
            message = "Unknown"

        return {"apiStatus": VolvoCarsValue(message)}

    async def async_get_brakes_status(
        self, vin: str = ""
    ) -> dict[str, VolvoCarsValueField | None]:
        """Get brakes status.

        Required scopes: openid conve:brake_status
        """
        return await self._async_get_field(_API_CONNECTED_ENDPOINT, "brakes", vin)

    async def async_get_command_accessibility(
        self, vin: str = ""
    ) -> dict[str, VolvoCarsValueField | None]:
        """Get availability status.

        Required scopes: openid conve:command_accessibility
        """
        return await self._async_get_field(
            _API_CONNECTED_ENDPOINT, "command-accessibility", vin
        )

    async def async_get_commands(
        self, vin: str = ""
    ) -> list[VolvoCarsAvailableCommand | None]:
        """Get available commands.

        Required scopes: openid conve:commands
        """
        body = await self._async_get(_API_CONNECTED_ENDPOINT, "commands", vin)
        items = self._get_data_list(body)
        return [VolvoCarsAvailableCommand.from_dict(item) for item in items]

    async def async_get_diagnostics(
        self, vin: str = ""
    ) -> dict[str, VolvoCarsValueField | None]:
        """Get diagnostics.

        Required scopes: openid conve:diagnostics_workshop
        """
        return await self._async_get_field(_API_CONNECTED_ENDPOINT, "diagnostics", vin)

    async def async_get_doors_status(
        self, vin: str = ""
    ) -> dict[str, VolvoCarsValueField | None]:
        """Get doors status.

        Required scopes: openid conve:doors_status conve:lock_status
        """
        return await self._async_get_field(_API_CONNECTED_ENDPOINT, "doors", vin)

    async def async_get_energy_capabilities(self, vin: str = "") -> dict[str, Any]:
        """Get energy state.

        Required scopes: openid energy:capability:read
        """
        return await self._async_get_data_dict(
            _API_ENERGY_V2_ENDPOINT, "capabilities", vin, data_key="getEnergyState"
        )

    async def async_get_energy_state(
        self, vin: str = ""
    ) -> dict[str, VolvoCarsValueStatusField | None]:
        """Get energy state.

        Required scopes: openid energy:state:read
        """
        body = await self._async_get(_API_ENERGY_V2_ENDPOINT, "state", vin)
        return {
            key: VolvoCarsValueStatusField.from_dict(value)
            for key, value in body.items()
        }

    async def async_get_engine_status(
        self, vin: str = ""
    ) -> dict[str, VolvoCarsValueField | None]:
        """Get engine status.

        Required scopes: openid conve:engine_status
        """
        return await self._async_get_field(
            _API_CONNECTED_ENDPOINT, "engine-status", vin
        )

    async def async_get_engine_warnings(
        self, vin: str = ""
    ) -> dict[str, VolvoCarsValueField | None]:
        """Get engine warnings.

        Required scopes: openid conve:diagnostics_engine_status
        """
        return await self._async_get_field(_API_CONNECTED_ENDPOINT, "engine", vin)

    async def async_get_fuel_status(
        self, vin: str = ""
    ) -> dict[str, VolvoCarsValueField | None]:
        """Get fuel status.

        Required scopes: openid conve:fuel_status conve:battery_charge_level
        """
        return await self._async_get_field(_API_CONNECTED_ENDPOINT, "fuel", vin)

    async def async_get_location(
        self, vin: str = ""
    ) -> dict[str, VolvoCarsLocation | None]:
        """Get location.

        Required scopes: openid location:read
        """
        data = await self._async_get_data_dict(_API_LOCATION_ENDPOINT, "location", vin)
        return {"location": VolvoCarsLocation.from_dict(data)}

    async def async_get_odometer(
        self, vin: str = ""
    ) -> dict[str, VolvoCarsValueField | None]:
        """Get odometer.

        Required scopes: openid conve:odometer_status
        """
        return await self._async_get_field(_API_CONNECTED_ENDPOINT, "odometer", vin)

    async def async_get_recharge_status(
        self, vin: str = ""
    ) -> dict[str, VolvoCarsValueField | None]:
        """Get recharge status.

        Required scopes: openid

        And at least one of:
            energy:recharge_status energy:battery_charge_level energy:electric_range
            energy:estimated_charging_time energy:charging_connection_status
            energy:charging_system_status energy:charging_current_limit
            energy:target_battery_level
        """
        return await self._async_get_field(_API_ENERGY_ENDPOINT, "recharge-status", vin)

    async def async_get_statistics(
        self, vin: str = ""
    ) -> dict[str, VolvoCarsValueField | None]:
        """Get statistics.

        Required scopes: openid conve:trip_statistics
        """
        return await self._async_get_field(_API_CONNECTED_ENDPOINT, "statistics", vin)

    async def async_get_tyre_states(
        self, vin: str = ""
    ) -> dict[str, VolvoCarsValueField | None]:
        """Get tyre states.

        Required scopes: openid conve:tyre_status
        """
        return await self._async_get_field(_API_CONNECTED_ENDPOINT, "tyres", vin)

    async def async_get_vehicles(self) -> list[str]:
        """Get vehicles.

        Required scopes: openid conve:vehicle_relation
        """
        url = f"{_API_URL}{_API_CONNECTED_ENDPOINT}"
        body = await self._async_request(hdrs.METH_GET, url, operation="vehicles")
        items = self._get_data_list(body)
        return [item["vin"] for item in items]

    async def async_get_vehicle_details(self, vin: str = "") -> VolvoCarsVehicle | None:
        """Get vehicle details.

        Required scopes: openid conve:vehicle_relation
        """
        data = await self._async_get_data_dict(_API_CONNECTED_ENDPOINT, "", vin)
        return VolvoCarsVehicle.from_dict(data)

    async def async_get_warnings(
        self, vin: str = ""
    ) -> dict[str, VolvoCarsValueField | None]:
        """Get warnings.

        Required scopes: openid conve:warnings
        """
        return await self._async_get_field(_API_CONNECTED_ENDPOINT, "warnings", vin)

    async def async_get_window_states(
        self, vin: str = ""
    ) -> dict[str, VolvoCarsValueField | None]:
        """Get window states.

        Required scopes: openid conve:windows_status
        """
        return await self._async_get_field(_API_CONNECTED_ENDPOINT, "windows", vin)

    async def async_execute_command(
        self, command: str, body: dict[str, Any] | None = None, vin: str = ""
    ) -> VolvoCarsCommandResult | None:
        """Execute a command.

        Required scopes: openid

        Per command:
        - engine-start: conve:engine_start_stop
        - engine-stop: conve:engine_start_stop
        - flash: conve:honk_flash
        - honk: conve:honk_flash
        - honk-flash: conve:honk_flash
        - lock: conve:lock
        - lock-reduced-guard: conve:lock
        - unlock: conve:unlock
        """
        body = await self._async_post(
            _API_CONNECTED_ENDPOINT, f"commands/{command}", body=body, vin=vin
        )
        data: dict = body.get("data", {})
        data["invoke_status"] = data.pop("invokeStatus", None)
        return VolvoCarsCommandResult.from_dict(data)

    async def async_get_access_token(self) -> str:
        """Get or refresh the access token."""
        operation = "token refresh"
        try:
            return await self._token_manager.async_get_access_token()
        except ClientResponseError as ex:
            _LOGGER.debug("Request [%s] error: %s", operation, ex.message)

            # Volvo returns a 400 for a stale refresh token. In such cases we also
            # want to raise a VolvoAuthException.
            if ex.status in (
                HTTPStatus.BAD_REQUEST,
                HTTPStatus.UNAUTHORIZED,
                HTTPStatus.FORBIDDEN,
            ):
                raise VolvoAuthException(ex.message, operation) from ex

            raise VolvoApiException(ex.message, operation) from ex
        except (ClientError, TimeoutError) as ex:
            _LOGGER.debug("Request [%s] error: %s", operation, ex.__class__.__name__)
            raise VolvoApiException(ex.__class__.__name__, operation) from ex

    async def _async_get_field(
        self, endpoint: str, operation: str, vin: str = ""
    ) -> dict[str, VolvoCarsValueField | None]:
        body = await self._async_get(endpoint, operation, vin)
        data: dict = body.get("data", {})
        return {
            key: VolvoCarsValueField.from_dict(value) for key, value in data.items()
        }

    async def _async_get_data_dict(
        self, endpoint: str, operation: str, vin: str = "", *, data_key: str = "data"
    ) -> dict[str, Any]:
        body = await self._async_get(endpoint, operation, vin)
        return cast(dict[str, Any], body.get(data_key, {}))

    async def _async_get(
        self, endpoint: str, operation: str, vin: str = ""
    ) -> dict[str, Any]:
        url = self._create_vin_url(endpoint, operation, vin)
        return await self._async_request(
            hdrs.METH_GET, url, operation=operation, vin=vin
        )

    async def _async_post(
        self,
        endpoint: str,
        operation: str,
        *,
        body: dict[str, Any] | None = None,
        vin: str = "",
    ) -> dict[str, Any]:
        url = self._create_vin_url(endpoint, operation, vin)
        return await self._async_request(
            hdrs.METH_POST, url, operation=operation, body=body, vin=vin
        )

    def _get_data_list(self, body: dict[str, Any]) -> list[Any]:
        return cast(list[Any], body.get("data", []))

    def _create_vin_url(self, endpoint: str, operation: str, vin: str = "") -> str:
        if not vin:
            vin = self.vin

        return (
            f"{_API_URL}{endpoint}/{vin}/{operation}"
            if operation
            else f"{_API_URL}{endpoint}/{vin}"
        )

    async def _async_request(
        self,
        method: str,
        url: str,
        *,
        operation: str,
        body: dict[str, Any] | None = None,
        vin: str = "",
    ) -> dict[str, Any]:
        access_token = await self.async_get_access_token()

        headers = {
            hdrs.AUTHORIZATION: f"Bearer {access_token}",
            "vcc-api-key": self.api_key,
        }

        if method == hdrs.METH_POST:
            headers[hdrs.CONTENT_TYPE] = "application/json"

        data: dict[str, Any] = {}

        try:
            _LOGGER.debug(
                "Request [%s]: %s %s",
                operation,
                method,
                redact_url(url, vin),
            )
            async with self._client.request(
                method, url, headers=headers, json=body, timeout=_API_REQUEST_TIMEOUT
            ) as response:
                _LOGGER.debug("Request [%s] status: %s", operation, response.status)
                json = await response.json()
                data = cast(dict[str, Any], json)
                _LOGGER.debug(
                    "Request [%s] response: %s",
                    operation,
                    redact_data(data, _DATA_TO_REDACT),
                )
                response.raise_for_status()
                return data
        except ClientResponseError as ex:
            if ex.status == HTTPStatus.NOT_FOUND:
                return {}

            _LOGGER.debug("Request [%s] error: %s", operation, ex.message)

            if ex.status == HTTPStatus.UNPROCESSABLE_ENTITY and "/commands" in url:
                return {
                    "data": {
                        "vin": vin,
                        "invokeStatus": "UNKNOWN",
                        "message": "",
                    }
                }

            redacted_exception = RedactedClientResponseError(ex, vin)
            message = redacted_exception.message

            if data and (error_data := data.get("error")):
                error = VolvoCarsErrorResult.from_dict(error_data)

                if error is not None:
                    message = f"{error.message} {error.description}".strip()

            if ex.status in (HTTPStatus.UNAUTHORIZED, HTTPStatus.FORBIDDEN):
                raise VolvoAuthException(message, operation) from redacted_exception

            raise VolvoApiException(message, operation) from redacted_exception

        except (ClientError, TimeoutError) as ex:
            _LOGGER.debug("Request [%s] error: %s", operation, ex.__class__.__name__)
            raise VolvoApiException(ex.__class__.__name__, operation) from ex


class RedactedClientResponseError(ClientResponseError):
    """Exception class that redacts sensitive data."""

    def __init__(self, exception: ClientResponseError, vin: str) -> None:
        """Initialize class."""

        redacted_url = self._redact_url(exception.request_info.url, vin)
        redacted_real_url = self._redact_url(exception.request_info.real_url, vin)
        redacted_request_info = RequestInfo(
            redacted_url,
            exception.request_info.method,
            exception.request_info.headers,
            redacted_real_url,
        )

        super().__init__(
            redacted_request_info,
            exception.history,
            status=exception.status,
            message=exception.message,
            headers=exception.headers,
        )

    def _redact_url(self, url: URL, vin: str) -> URL:
        redacted_url = redact_url(str(url), vin)
        return URL(redacted_url)



"""Volvo Cars API models."""

from dataclasses import KW_ONLY, dataclass, field, is_dataclass
from datetime import datetime
import inspect
import re
from typing import Any, Self, TypeVar

T = TypeVar("T", bound="VolvoCarsApiBaseModel")
_TO_SNAKE_CASE_REGEX = re.compile(r"(?<=[a-z0-9])([A-Z])")


def _sanitize_json_key(key: str) -> str:
    key = "description" if key == "descriptions" else key
    key = _TO_SNAKE_CASE_REGEX.sub(r"_\1", key)
    return key.lower()


@dataclass
class VolvoCarsApiBaseModel:
    """Base API model."""

    _: KW_ONLY
    extra_data: dict[str, Any] = field(default_factory=dict[str, Any])

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self | None:
        """Create instance from json dict."""
        parameters = inspect.signature(cls).parameters
        class_data: dict[str, Any] = {}
        extra_data: dict[str, Any] = {}

        for key, value in data.items():
            key = _sanitize_json_key(key)
            if key in parameters:
                # Check if the field is a dataclass and the value is a dict
                param_type = parameters[key].annotation
                if (
                    is_dataclass(param_type)
                    and isinstance(param_type, type)
                    and issubclass(param_type, VolvoCarsApiBaseModel)
                    and isinstance(value, dict)
                ):
                    # Recursive call for nested dataclasses
                    class_data[key] = param_type.from_dict(value)
                elif key in ("timestamp", "updatedAt") and isinstance(value, str):
                    if value:
                        class_data["timestamp"] = datetime.fromisoformat(value)
                else:
                    class_data[key] = value
            else:
                extra_data[key] = value

        if len(class_data) == 0:
            return None

        class_data["extra_data"] = extra_data
        return cls(**class_data)

    def get(self, key: str) -> Any:
        """Get a specific key from the API field."""
        return self.extra_data.get(key)


@dataclass
class VolvoCarsModel(VolvoCarsApiBaseModel):
    """Representation of a Volvo Cars model."""

    model: str
    steering: str
    upholstery: str | None = None


@dataclass
class VolvoCarsImages(VolvoCarsApiBaseModel):
    """Representation of Volvo Cars images."""

    exterior_image_url: str
    internal_image_url: str


@dataclass
class VolvoCarsVehicle(VolvoCarsApiBaseModel):
    """Representation of a Volvo Cars vehicle."""

    vin: str
    model_year: int
    gearbox: str
    fuel_type: str
    images: VolvoCarsImages
    description: VolvoCarsModel
    external_colour: str | None = None
    battery_capacity_kwh: float | None = None

    def has_battery_engine(self) -> bool:
        """Determine if vehicle has a battery engine."""
        return self.fuel_type in ("ELECTRIC", "PETROL/ELECTRIC", "NONE")

    def has_combustion_engine(self) -> bool:
        """Determine if vehicle has a combustion engine."""
        return self.fuel_type in ("DIESEL", "PETROL", "PETROL/ELECTRIC")


@dataclass
class VolvoCarsValue(VolvoCarsApiBaseModel):
    """API value model."""

    value: Any


@dataclass
class VolvoCarsValueField(VolvoCarsValue):
    """API value field model."""

    timestamp: datetime | None = None
    unit: str | None = None


@dataclass
class VolvoCarsValueStatusField(VolvoCarsValueField):
    """API value and status field model."""

    value: Any = None
    status: str | None = None


@dataclass
class VolvoCarsGeometry(VolvoCarsApiBaseModel):
    """API geometry model."""

    coordinates: list[float] = field(default_factory=list[float])


@dataclass
class VolvoCarsLocationProperties(VolvoCarsApiBaseModel):
    """API location properties model."""

    heading: str
    timestamp: datetime | None = None


@dataclass
class VolvoCarsLocation(VolvoCarsApiBaseModel):
    """API location model."""

    type: str
    properties: VolvoCarsLocationProperties
    geometry: VolvoCarsGeometry


@dataclass
class VolvoCarsAvailableCommand(VolvoCarsApiBaseModel):
    """Available command model."""

    command: str
    href: str


@dataclass
class VolvoCarsCommandResult(VolvoCarsApiBaseModel):
    """Command result model."""

    vin: str
    invoke_status: str
    message: str


@dataclass
class TokenResponse(VolvoCarsApiBaseModel):
    """Authorization response model."""

    access_token: str
    refresh_token: str
    token_type: str
    expires_in: int
    expires_at: float
    id_token: str | None = None


@dataclass
class VolvoCarsErrorResult(VolvoCarsApiBaseModel):
    """Error result model."""

    message: str | None = None
    description: str | None = None


class VolvoApiException(Exception):
    """Thrown when an API request fails."""

    def __init__(self, message: str = "", operation: str = "") -> None:
        """Initialize exception."""
        self.message = message
        self.operation = operation


class VolvoAuthException(VolvoApiException):
    """Thrown when the authentication fails."""



"""Volvo Cars API utils."""

from collections.abc import Iterable, Mapping
from typing import Any

REDACTED = "**REDACTED**"


def redact_data(data: Mapping, to_redact: Iterable[Any]) -> dict:
    """Redact sensitive data in a dict."""

    redacted = {**data}

    for key, value in redacted.items():
        if value is None:
            continue
        if isinstance(value, str) and not value:
            continue
        if key in to_redact:
            redacted[key] = REDACTED
        elif isinstance(value, Mapping):
            redacted[key] = redact_data(value, to_redact)
        elif isinstance(value, list):
            redacted[key] = [redact_data(item, to_redact) for item in value]

    return redacted


def redact_url(url: str, vin: str) -> str:
    """Redact VIN from URL."""
    return url.replace(vin, REDACTED) if vin else url



"""Volvo Cars Auth API."""

from abc import ABC, abstractmethod
from asyncio import Lock
import base64
import hashlib
import logging
import secrets
import time
from typing import Any, cast

from aiohttp import ClientError, ClientSession, ClientTimeout, hdrs
from yarl import URL

from .models import TokenResponse, VolvoApiException, VolvoAuthException
from .util import redact_data

AUTHORIZE_URL = "https://volvoid.eu.volvocars.com/as/authorization.oauth2"
TOKEN_URL = "https://volvoid.eu.volvocars.com/as/token.oauth2"

_LOGGER = logging.getLogger(__name__)

_API_REQUEST_TIMEOUT = ClientTimeout(total=30)
_CLOCK_OUT_OF_SYNC_MAX_SEC = 20
_DATA_TO_REDACT = [
    "access_token",
    "code",
    "id",
    "id_token",
    "href",
    "refresh_token",
    "target",
    "username",
]


def _generate_code_verifier(code_verifier_length: int = 128) -> str:
    if not 43 <= code_verifier_length <= 128:
        msg = (
            "Parameter `code_verifier_length` must validate"
            "`43 <= code_verifier_length <= 128`."
        )
        raise ValueError(msg)
    return secrets.token_urlsafe(96)[:code_verifier_length]


def _compute_code_challenge(code_verifier: str) -> str:
    if not 43 <= len(code_verifier) <= 128:
        msg = (
            "Parameter `code_verifier` must validate `43 <= len(code_verifier) <= 128`."
        )
        raise ValueError(msg)

    hashed = hashlib.sha256(code_verifier.encode("ascii")).digest()
    encoded = base64.urlsafe_b64encode(hashed)
    return encoded.decode("ascii").replace("=", "")


class AccessTokenManager(ABC):
    """Access Token manager."""

    def __init__(self, websession: ClientSession) -> None:
        """Initialize the auth."""
        self.websession = websession

    @abstractmethod
    async def async_get_access_token(self) -> str:
        """Return a valid access token."""


class VolvoCarsAuth(AccessTokenManager):
    """Volvo Cars authentication."""

    def __init__(
        self,
        websession: ClientSession,
        *,
        client_id: str,
        client_secret: str,
        scopes: list[str],
        redirect_uri: str,
        code_verifier_length: int = 128,
    ) -> None:
        """Initialize the auth."""
        super().__init__(websession)

        self._client_id = client_id
        self._client_secret = client_secret
        self._scopes = scopes
        self._redirect_uri = redirect_uri

        self._code_verifier = _generate_code_verifier(code_verifier_length)

        credentials = f"{self._client_id}:{self._client_secret}"
        self._encoded_credentials = base64.b64encode(
            credentials.encode("utf-8")
        ).decode("utf-8")

        self._token: TokenResponse
        self._token_lock = Lock()

    @property
    def token(self) -> TokenResponse | None:
        """Get current token."""
        return self._token

    @property
    def valid_token(self) -> bool:
        """Return if token is still valid."""
        return (
            self._token.expires_at > time.time() + _CLOCK_OUT_OF_SYNC_MAX_SEC
            if self._token
            else False
        )

    async def async_get_access_token(self) -> str:
        """Return a valid access token, refresh if needed."""
        await self.async_ensure_token_valid()
        return self._token.access_token

    async def async_ensure_token_valid(self) -> None:
        """Ensure that the current token is valid."""
        async with self._token_lock:
            if self.valid_token:
                return

            if not self._token:
                raise ValueError("No token available.")

            await self.async_refresh_token(self._token.refresh_token)

    def get_auth_uri(self, state: str | None = None) -> str:
        """Get the full authorization URL."""
        query = {
            "response_type": "code",
            "client_id": self._client_id,
            "redirect_uri": self._redirect_uri,
            "scope": " ".join(self._scopes),
            "code_challenge": _compute_code_challenge(self._code_verifier),
            "code_challenge_method": "S256",
        }

        if state:
            query |= {"state": state}

        return str(URL(AUTHORIZE_URL).with_query(query))

    async def async_request_token(self, code: str) -> TokenResponse:
        """Request access token."""
        data = {
            "grant_type": "authorization_code",
            "code": code,
            "redirect_uri": self._redirect_uri,
            "code_verifier": self._code_verifier,
        }

        response = await self._async_request(
            hdrs.METH_POST,
            TOKEN_URL,
            headers={"Authorization": f"Basic {self._encoded_credentials}"},
            data=data,
            operation="tokens",
        )

        self._token = self._create_token_response(response)
        return self._token

    async def async_refresh_token(self, refresh_token: str) -> TokenResponse:
        """Refresh token."""
        data = {
            "grant_type": "refresh_token",
            "refresh_token": refresh_token,
        }

        response = await self._async_request(
            hdrs.METH_POST,
            TOKEN_URL,
            headers={"Authorization": f"Basic {self._encoded_credentials}"},
            data=data,
            operation="token refresh",
        )

        self._token = self._create_token_response(response)
        return self._token

    def _create_token_response(self, token: dict[str, Any]) -> TokenResponse:
        token["expires_in"] = int(token["expires_in"])
        token["expires_at"] = time.time() + token["expires_in"]

        token_response = TokenResponse.from_dict(token)

        if token_response is None:
            raise ValueError("Could not create token response.")

        return token_response

    async def _async_request(
        self,
        method: str,
        url: str,
        *,
        params: dict[str, str] | None = None,
        headers: dict[str, str] | None = None,
        data: dict[str, str] | None = None,
        json: dict[str, Any] | None = None,
        operation: str = "",
    ) -> dict[str, Any]:
        _LOGGER.debug("Request [%s]", operation)

        try:
            async with self.websession.request(
                method,
                url,
                params=params,
                headers=headers,
                data=data,
                json=json,
                timeout=_API_REQUEST_TIMEOUT,
            ) as response:
                _LOGGER.debug("Request [%s] status: %s", operation, response.status)

                json = await response.json()
                data = cast(dict[str, Any], json)

                _LOGGER.debug(
                    "Request [%s] response: %s",
                    operation,
                    redact_data(data, _DATA_TO_REDACT),
                )

                response.raise_for_status()
                return data

        except ClientError as ex:
            _LOGGER.debug("Request [%s] error: %s", operation, ex.__class__.__name__)
            raise VolvoAuthException(ex.__class__.__name__, operation) from ex
        except TimeoutError as ex:
            _LOGGER.debug("Request [%s] error: %s", operation, ex.__class__.__name__)
            raise VolvoApiException(ex.__class__.__name__, operation) from ex



"""Volvo Cars API scopes."""

from enum import StrEnum


class RestrictedScope(StrEnum):
    """Privacy and security related scopes."""

    LOCK = "conve:lock"
    UNLOCK = "conve:unlock"
    ENGINE_START_STOP = "conve:engine_start_stop"
    HONK_FLASH = "conve:honk_flash"
    LOCATION = "location:read"

DEFAULT_SCOPES = [
    "openid",
    "conve:battery_charge_level",
    "conve:brake_status",
    "conve:climatization_start_stop",
    "conve:command_accessibility",
    "conve:commands",
    "conve:diagnostics_engine_status",
    "conve:diagnostics_workshop",
    "conve:doors_status",
    "conve:engine_status",
    "conve:fuel_status",
    "conve:lock_status",
    "conve:odometer_status",
    "conve:trip_statistics",
    "conve:tyre_status",
    "conve:vehicle_relation",
    "conve:warnings",
    "conve:windows_status",
    "energy:capability:read",
    "energy:state:read"
]

DEPRECATED_SCOPES = [
    "energy:battery_charge_level",
    "energy:charging_connection_status",
    "energy:charging_current_limit",
    "energy:charging_system_status",
    "energy:electric_range",
    "energy:estimated_charging_time",
    "energy:recharge_status",
    "energy:target_battery_level",
]

ALL_SCOPES = DEFAULT_SCOPES + [s.value for s in RestrictedScope]
